import jax.numpy as jnp
from .utils import sigmoid
from .utils import log1pexp

def compute_W_loss(W, batch_X, batch_H, l1_W):

    k, d = W.shape
    t, _ = batch_X.shape

    W_pos = log1pexp(W)
    batch_H_pos = log1pexp(batch_H)
    
    reconstruction = batch_H_pos @ W_pos

    reconstruction_loss = jnp.linalg.norm(reconstruction - batch_X)/t/d

    l1_loss = jnp.abs(W_pos).mean() * l1_W

    return reconstruction_loss + l1_loss


def compute_batch_H_loss(batch_H, batch_X, W, l1_W):

    k, d = W.shape
    t, _ = batch_X.shape
    
    W_pos = log1pexp(W)
    batch_H_pos = log1pexp(batch_H)
    
    reconstruction = batch_H_pos @ W_pos

    reconstruction_loss = jnp.linalg.norm(reconstruction - batch_X)/t/d

    l1_loss = jnp.abs(W_pos).mean() * l1_W

    return reconstruction_loss + l1_loss



import jax.numpy as jnp


log1pexp = lambda x: jnp.log1p(jnp.exp(x))
logexpm1 = lambda x: jnp.log(jnp.expm1(x))


def sigmoid(Z):
    A=1/(1+(jnp.exp((-Z))))
    return A


def order_factors(H, W):
    import numpy as np
    from numpy import argsort
    Hnorm = np.linalg.norm(H, axis = 0)
    Wnorm = np.linalg.norm(W, axis = 1)
    # W /= Wnorm[:,None]
    # H *= Wnorm[None]
    Knorm = Hnorm*Wnorm
    order = argsort(Knorm)[::-1]
    H = H[:,order]
    W = W[order]
    return H, W


def get_subfolders(folder, extension = '*', verbose = False, make_plot_folder = False, windows = False):
    from glob import glob
    import os
    ''' extract subfolders and paths of a main directory containing extension'''
    
    folder_paths = list(filter(lambda v: os.path.isdir(v), sorted(glob(folder+ extension))))
    if windows == True:  div = '\\'
    else: div = '/'
    folders = list(map(lambda v: v.split(div)[-1], folder_paths))
    dirs = {}
    dirs['main'] = folder
    for i in range(len(folders)):
        if os.path.isdir(folder_paths[i]):
            dirs[folders[i]] = folder_paths[i] + div
    if verbose:
        [print(k) for ind, k in enumerate(dirs.keys())]
    return dirs


import jax
import jax.numpy as jnp
from jax import jit
from jax import random
import numpy as np
from time import time
import optax
from sklearn.decomposition._nmf import _initialize_nmf as initialize_nmf

from .utils import sigmoid, log1pexp, order_factors
from .parameters import Log
from .initialize import initialize
from .losses import compute_batch_H_loss
from .losses import compute_W_loss
from .updates import update_W_step
from .updates import update_W_batch_H_step


def nmf(X, k, parameters, print_iter = 20, initial_values = None, save_iter = None, save_path = None):
    t, d = X.shape
    X_min = np.min(X)

    if X_min<0:
        print('Error! X must be positive :)')
        raise ValueError
        
    if parameters.batch_size>t:
        parameters.batch_size = t
        print('batch size greater than t - resetting batch size')
   
    if initial_values is None:
        print('intializing values with nnsvd')
        t0 = time()
        H, W = initialize_nmf(X, k, 'nndsvd')
        t1 = time()
        print(f'time: {np.round((t1-t0)/60,4)}mins')
    else:
        H = initial_values['H']
        W = initial_values['W']
        print('H & W initialized with given initial values')


    log = Log()

    ### optimizers
    optimizer_W = optax.adam(learning_rate = parameters.step_size)
    opt_state_W = optimizer_W.init(W)

    optimizer_H = optax.adam(learning_rate = parameters.step_size)
    opt_state_H = optimizer_H.init(H)

    update_W_batch_H_step_jit = jax.jit(update_W_batch_H_step, static_argnames=['optimizer_W', 'total_batch_num', 'parameters'])
    

    H_prev = np.copy(H)
    total_batch_num = np.int(np.round(t/parameters.batch_size))
    print(f'total batch num: {total_batch_num}')
    
    t0 = time()

    key = random.PRNGKey(42)
    try:
        for i in range(parameters.max_iter):
            key, shuffle_key = random.split(key)
            
            W, opt_state_W, grad_H, loss_batch = update_W_batch_H_step_jit(X, H, W, optimizer_W, opt_state_W, opt_state_H, parameters, total_batch_num, shuffle_key)
            loss_batch = float(loss_batch) 
            grad_norm = np.linalg.norm(grad_H)
            log.total_loss.append(loss_batch)
            log.grad_norm.append(grad_norm)
            
            updates_H, opt_state_H = optimizer_H.update(grad_H, opt_state_H, H)
            H = optax.apply_updates(H, updates_H)
            H_diff = np.linalg.norm(H-H_prev)/k/t*1e5
            H_prev = np.copy(H)

            t1 = time()
            tdiff = np.round(t1-t0,2)/60

            statement = f"Iteration {i}, loss={np.round(loss_batch,10)}, h_diff={np.round(H_diff,7)}, grad_norm={np.round(grad_norm,10)}, time={np.round(tdiff,4)}min" 

            if i % print_iter == 0:    
                print(statement)
                print('log1pexp')
                print(log1pexp(H).max(), log1pexp(W).max())

            if save_iter is not None:
                if i % save_iter == 0:
                    H_, W_ = order_factors(H, W)
                    results = {}
                    results['H'] = log1pexp(H_)
                    results['W'] = log1pexp(W_)
                    results['ite'] = i 
                    results['loss'] = log.total_loss
                    np.save(save_path + f'_ite_{i:05d}.npy', results)
                    print('intermediate results saved')

            if loss_batch < parameters.min_loss:
                print(statement)
                print("Fitting converged!")
                break
            if i>10:
                if H_diff < parameters.min_diff:
                    print(statement)
                    print("Fitting converged! Minimal difference reached.")
                    break
        if i == parameters.max_iter-1:
            print(statement)
            print('Reached maximum number of set iterations!')
        else:
            print(f'reached iteration {i}')

    except KeyboardInterrupt:
        print('caught Keyboard interrupt')

    ### project solution back into feasible space
    H = log1pexp(H)
    W = log1pexp(W)

    ### normalize and order factors by norm
    H, W = order_factors(H, W)
    
    return np.array(H), np.array(W), log








import jax.numpy as jnp
import numpy as np
from jax.tree_util import register_pytree_node_class


@register_pytree_node_class
class Parameters():
    def __init__(self, l1_W=0.0, max_iter=10000, min_loss=1e-5, batch_size=20, step_size=1e-2, min_diff = 1e-7):
        self.l1_W = l1_W
        self.max_iter= max_iter
        self.min_loss = min_loss
        self.batch_size = batch_size
        self.step_size = step_size
        self.min_diff = min_diff

    def tree_flatten(self):
        children = (self.l1_W, self.max_iter, self.min_loss, self.batch_size, self.step_size, self.min_diff)
        aux_data = None
        return (children, aux_data)
    
    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)


class Log():
    def __init__(self,):
        self.l1_loss_W = []
        self.reconstruction_loss = []
        self.total_loss = []
        self.grad_norm = []




import optax
import jax
from .losses import compute_W_loss
from .losses import compute_batch_H_loss
from jax import random
import jax.numpy as jnp


def update_W_step(W, optimizer_W, opt_state_W, batch_X, batch_H, l1_W):
    loss_W, grad_W = jax.value_and_grad(compute_W_loss)(W, batch_X, batch_H, l1_W)
    updates, opt_state_W = optimizer_W.update(grad_W, opt_state_W, W)
    W = optax.apply_updates(W, updates)
    return W, opt_state_W, loss_W


def update_W_batch_H_step(
    X,
    H,
    W,
    optimizer_W,
    opt_state_W,
    opt_state_H,
    parameters,
    total_batch_num,
    shuffle_key,
):
    print("compiling update function")
    t, d = X.shape

    shuffled_indices = jnp.arange(t)
    random.permutation(shuffle_key, shuffled_indices)

    grad_H_batches = []
    for j in range(total_batch_num):
        batch_indices = shuffled_indices[
            j * parameters.batch_size : (j + 1) * parameters.batch_size
        ]
        batch_X = X[batch_indices]
        batch_H = H[batch_indices]

        grad_H_batch = jax.grad(compute_batch_H_loss)(
            batch_H, batch_X, W, parameters.l1_W
        )

        W, opt_state_W, loss_batch = update_W_step(
            W, optimizer_W, opt_state_W, batch_X, batch_H, parameters.l1_W
        )
        grad_H_batches.append(grad_H_batch)

    grad_H_batches = jnp.vstack(grad_H_batches)
    grad_H = grad_H_batches[jnp.argsort(shuffled_indices).squeeze()]

    return W, opt_state_W, grad_H, loss_batch



from setuptools import setup

setup(
        name='nmfx',
        version='0.1',
        description='Stochastic Non-Negative Matrix Factorization',
        url='https://github.com/vrutten/nmfx',
        author='Virginia MS Rutten, Jan Funke',
        author_email='ruttenv@janelia.hhmi.org, funkej@janelia.hhmi.org',
        license='MIT',
        packages=[
            'nmfx'
        ],
        install_requires=[
            'numpy',
            'jax',
            'tqdm'
        ],
        scripts=[
            'scripts/nmfx',
            'scripts/nmfx_hdf5'
        ]
)


import jax.numpy as jnp
from .utils import sigmoid, log1pexp
import numpy as np

def initialize(X, k, how = 'random'):
    t, d = X.shape
    if how == 'random':
        H = log1pexp(np.random.randn(t, k) + 2)
        W = log1pexp(np.random.randn(k, d) + 2)
    return H, W
